"""Wordnet lexicon validation.

This module is for checking whether the the contents of a lexicon are
valid according to a series of checks. Those checks are:

====  ==========================================================
Code  Message
====  ==========================================================
E101  ID is not unique within the lexicon.
W201  Lexical entry has no senses.
W202  Redundant sense between lexical entry and synset.
W203  Redundant lexical entry with the same lemma and synset.
E204  Synset of sense is missing.
W301  Synset is empty (not associated with any lexical entries).
W302  ILI is repeated across synsets.
W303  Proposed ILI is missing a definition.
W304  Existing ILI has a spurious definition.
E401  Relation target is missing or invalid.
W402  Relation type is invalid for the source and target.
W403  Redundant relation between source and target.
W404  Reverse relation is missing.
W501  Synset's part-of-speech is different from its hypernym's.
W502  Relation is a self-loop.
====  ==========================================================

"""

from typing import (
    Sequence,
    Iterator,
    Tuple,
    List,
    Dict,
    Optional,
    Type,
    Union,
    Callable,
    cast,
)
from collections import Counter
from itertools import chain

from wn import lmf
from wn.constants import (
    SENSE_RELATIONS,
    SENSE_SYNSET_RELATIONS,
    SYNSET_RELATIONS,
    REVERSE_RELATIONS,
)
from wn.util import ProgressHandler, ProgressBar


_Ids = Dict[str, Counter]
_Result = Dict[str, Dict]
_CheckFunction = Callable[[lmf.Lexicon, _Ids], _Result]
_Report = Dict[str, Dict[str, Union[str, _Result]]]


def _non_unique_id(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """ID is not unique within the lexicon"""
    return _multiples(
        chain(
            [lex["id"]],
            (f["id"] for e in _entries(lex) for f in _forms(e) if f.get("id")),
            (sb["id"] for sb in lex.get("frames", []) if sb.get("id")),
            ids["entry"],
            ids["sense"],
            ids["synset"],
        )
    )


def _has_no_senses(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """lexical entry has no senses"""
    return {e["id"]: {} for e in _entries(lex) if not _senses(e)}


def _redundant_sense(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """redundant sense between lexical entry and synset"""
    result: _Result = {}
    for e in _entries(lex):
        redundant = _multiples(s["synset"] for s in _senses(e))
        result.update(
            (s["id"], {"entry": e["id"], "synset": s["synset"]})
            for s in _senses(e)
            if s["synset"] in redundant
        )
    return result


def _redundant_entry(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """redundant lexical entry with the same lemma and synset"""
    redundant = _multiples(
        (e["lemma"]["writtenForm"], s["synset"])
        for e in _entries(lex)
        for s in _senses(e)
    )
    return {form: {"synset": synset} for form, synset in redundant}


def _missing_synset(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """synset of sense is missing"""
    synset_ids = ids["synset"]
    return {
        s["id"]: {"synset": s["synset"]}
        for e in _entries(lex)
        for s in _senses(e)
        if s["synset"] not in synset_ids
    }


def _empty_synset(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """synset is empty (not associated with any lexical entries)"""
    synsets = {s["synset"] for e in _entries(lex) for s in _senses(e)}
    return {ss["id"]: {} for ss in _synsets(lex) if ss["id"] not in synsets}


def _repeated_ili(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """ILI is repeated across synsets"""
    repeated = _multiples(
        ss["ili"] for ss in _synsets(lex) if ss["ili"] and ss["ili"] != "in"
    )
    return {
        ss["id"]: {"ili": ss["ili"]} for ss in _synsets(lex) if ss["ili"] in repeated
    }


def _missing_ili_definition(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """proposed ILI is missing a definition"""
    return {
        ss["id"]: {}
        for ss in _synsets(lex)
        if ss["ili"] == "in" and not ss.get("ili_definition")
    }


def _spurious_ili_definition(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """existing ILI has a spurious definition"""
    return {
        ss["id"]: {"ili_definitin": ss["ili_definition"]}
        for ss in _synsets(lex)
        if ss["ili"] and ss["ili"] != "in" and ss.get("ili_definition")
    }


def _missing_relation_target(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """relation target is missing or invalid"""
    result = {
        s["id"]: {"type": r["relType"], "target": r["target"]}
        for s, r in _sense_relations(lex)
        if r["target"] not in ids["sense"] and r["target"] not in ids["synset"]
    }
    result.update(
        (ss["id"], {"type": r["relType"], "target": r["target"]})
        for ss, r in _synset_relations(lex)
        if r["target"] not in ids["synset"]
    )
    return result


def _invalid_relation_type(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """relation type is invalid for the source and target"""
    result = {
        s["id"]: {"type": r["relType"], "target": r["target"]}
        for s, r in _sense_relations(lex)
        if (r["target"] in ids["sense"] and r["relType"] not in SENSE_RELATIONS)
        or (r["target"] in ids["synset"] and r["relType"] not in SENSE_SYNSET_RELATIONS)
    }
    result.update(
        (ss["id"], {"type": r["relType"], "target": r["target"]})
        for ss, r in _synset_relations(lex)
        if r["relType"] not in SYNSET_RELATIONS
    )
    return result


def _redundant_relation(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """redundant relation between source and target"""
    redundant = _multiples(
        chain(
            ((s["id"], r["relType"], r["target"]) for s, r in _sense_relations(lex)),
            ((ss["id"], r["relType"], r["target"]) for ss, r in _synset_relations(lex)),
        )
    )
    return {src: {"type": typ, "target": tgt} for src, typ, tgt in redundant}


def _missing_reverse_relation(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """reverse relation is missing"""
    regular = {
        (s["id"], r["relType"], r["target"])
        for s, r in _sense_relations(lex)
        if r["target"] in ids["sense"]
    }
    regular.update(
        (ss["id"], r["relType"], r["target"]) for ss, r in _synset_relations(lex)
    )
    return {
        tgt: {"type": REVERSE_RELATIONS[typ], "target": src}
        for src, typ, tgt in regular
        if typ in REVERSE_RELATIONS
        and (tgt, REVERSE_RELATIONS[typ], src) not in regular
    }


def _hypernym_wrong_pos(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """synset's part-of-speech is different from its hypernym's"""
    sspos = {ss["id"]: ss.get("partOfSpeech") for ss in _synsets(lex)}
    return {
        ss["id"]: {"type": r["relType"], "target": r["target"]}
        for ss, r in _synset_relations(lex)
        if r["relType"] == "hypernym" and ss.get("partOfSpeech") != sspos[r["target"]]
    }


def _self_loop(lex: lmf.Lexicon, ids: _Ids) -> _Result:
    """relation is a self-loop"""
    relations = chain(_sense_relations(lex), _synset_relations(lex))
    return {
        x["id"]: {"type": r["relType"], "target": r["target"]}
        for x, r in relations
        if x["id"] == r["target"]
    }


# Helpers


def _multiples(iterable):
    counts = Counter(iterable)
    return {x: {"count": cnt} for x, cnt in counts.items() if cnt > 1}


def _entries(lex: lmf.Lexicon) -> List[lmf.LexicalEntry]:
    return lex.get("entries", [])


def _forms(e: lmf.LexicalEntry) -> List[lmf.Form]:
    return e.get("forms", [])


def _senses(e: lmf.LexicalEntry) -> List[lmf.Sense]:
    return e.get("senses", [])


def _synsets(lex: lmf.Lexicon) -> List[lmf.Synset]:
    return lex.get("synsets", [])


def _sense_relations(lex: lmf.Lexicon) -> Iterator[Tuple[lmf.Sense, lmf.Relation]]:
    for e in _entries(lex):
        for s in _senses(e):
            for r in s.get("relations", []):
                yield (s, r)


def _synset_relations(lex: lmf.Lexicon) -> Iterator[Tuple[lmf.Synset, lmf.Relation]]:
    for ss in _synsets(lex):
        for r in ss.get("relations", []):
            yield (ss, r)


# Check codes and messages
#
# categories:
#   E - errors
#   W - warnings
# subcategories:
#   100 - general
#   200 - words and senses
#   300 - synsets and ilis
#   400 - relations
#   500 - graph and taxonomy

_codes: Dict[str, _CheckFunction] = {
    # 100 - general
    "E101": _non_unique_id,
    # 200 - words and senses
    "W201": _has_no_senses,
    "W202": _redundant_sense,
    "W203": _redundant_entry,
    "E204": _missing_synset,
    # 300 - synsets and ilis
    "W301": _empty_synset,
    "W302": _repeated_ili,
    "W303": _missing_ili_definition,
    "W304": _spurious_ili_definition,
    # 400 - relations
    "E401": _missing_relation_target,
    "W402": _invalid_relation_type,
    "W403": _redundant_relation,
    "W404": _missing_reverse_relation,
    # 500 - graph
    "W501": _hypernym_wrong_pos,
    "W502": _self_loop,
}


def _select_checks(select: Sequence[str]) -> List[Tuple[str, _CheckFunction, str]]:
    selectset = set(select)
    return [
        (code, func, func.__doc__ or "")
        for code, func in _codes.items()
        if code in selectset or code[0] in selectset
    ]


# Main function


def validate(
    lex: Union[lmf.Lexicon, lmf.LexiconExtension],
    select: Sequence[str] = ("E", "W"),
    progress_handler: Optional[Type[ProgressHandler]] = ProgressBar,
) -> _Report:
    """Check *lex* for validity and return a report of the results.

    The *select* argument is a sequence of check codes (e.g.,
    ``E101``) or categories (``E`` or ``W``).

    The *progress_handler* parameter takes a subclass of
    :class:`wn.util.ProgressHandler`. An instance of the class will be
    created, used, and closed by this function.
    """
    if lex.get("extends"):
        print("validation of lexicon extensions is not supported")
        return {}
    lex = cast(lmf.Lexicon, lex)

    if progress_handler is None:
        progress_handler = ProgressHandler

    ids: _Ids = {
        "entry": Counter(entry["id"] for entry in _entries(lex)),
        "sense": Counter(
            sense["id"] for entry in _entries(lex) for sense in _senses(entry)
        ),
        "synset": Counter(synset["id"] for synset in _synsets(lex)),
    }

    checks = _select_checks(select)

    progress = progress_handler(message="Validate", total=len(checks))

    report: _Report = {}
    for code, func, message in checks:
        progress.set(status=func.__name__.replace("_", " "))
        report[code] = {"message": message, "items": func(lex, ids)}
        progress.update()
    progress.set(status="")
    progress.close()
    return report



import nox


@nox.session
def lint(session):
    session.install(".[test,web]")
    session.run("flake8", "--max-line-length", "88", "wn", "tests")
    session.run("mypy", "--python-version", "3.7", "wn")


@nox.session
def test(session):
    session.install("pytest")
    session.install(".")
    session.run("pytest")



import pytest

import wn


@pytest.mark.usefixtures("mini_db")
def test_word_derived_words():
    assert len(wn.word("test-en-example-n").derived_words()) == 1
    assert len(wn.word("test-es-ejemplo-n").derived_words()) == 1


@pytest.mark.usefixtures("mini_db")
def test_synset_hypernyms():
    assert wn.synset("test-en-0002-n").hypernyms() == [wn.synset("test-en-0001-n")]
    assert wn.synset("test-en-0001-n").hypernyms() == []


@pytest.mark.usefixtures("mini_db")
def test_synset_hypernyms_expand_default():
    assert wn.synset("test-es-0002-n").hypernyms() == [wn.synset("test-es-0001-n")]
    assert wn.synset("test-es-0001-n").hypernyms() == []


@pytest.mark.usefixtures("mini_db")
def test_synset_hypernyms_expand_empty():
    w = wn.Wordnet(lang="es", expand="")
    assert w.synset("test-es-0002-n").hypernyms() == []


@pytest.mark.usefixtures("mini_db")
def test_synset_hypernyms_expand_specified():
    w = wn.Wordnet(lang="es", expand="test-en")
    assert w.synset("test-es-0002-n").hypernyms() == [w.synset("test-es-0001-n")]


@pytest.mark.usefixtures("mini_db")
def test_synset_relations():
    w = wn.Wordnet(lang="en")
    assert w.synset("test-en-0002-n").relations() == {
        "hypernym": [w.synset("test-en-0001-n")],
        "hyponym": [w.synset("test-en-0004-n")],
    }


@pytest.mark.usefixtures("mini_db")
def test_sense_get_related():
    w = wn.Wordnet("test-en")
    assert w.sense("test-en-example-n-0002-01").get_related() == [
        w.sense("test-en-exemplify-v-0003-01")
    ]


@pytest.mark.usefixtures("mini_db")
def test_sense_relations():
    w = wn.Wordnet("test-en")
    assert w.sense("test-en-example-n-0002-01").relations() == {
        "derivation": [w.sense("test-en-exemplify-v-0003-01")]
    }


@pytest.mark.usefixtures("mini_db_1_1")
def test_extension_relations():
    # default mode
    assert wn.synset("test-en-0007-v").hypernyms() == [wn.synset("test-en-ext-0009-v")]
    assert wn.synset("test-en-ext-0009-v").hyponyms() == [wn.synset("test-en-0007-v")]
    assert wn.sense("test-en-information-n-0001-01").get_related("pertainym") == [
        wn.sense("test-en-ext-info-n-0001-01")
    ]
    assert wn.sense("test-en-ext-info-n-0001-01").get_related("pertainym") == [
        wn.sense("test-en-information-n-0001-01")
    ]

    # restricted to base
    w = wn.Wordnet(lexicon="test-en")
    assert w.synset("test-en-0007-v").hypernyms() == []
    assert w.sense("test-en-information-n-0001-01").get_related("pertainym") == []

    # base and extension
    w = wn.Wordnet(lexicon="test-en test-en-ext")
    assert w.synset("test-en-0007-v").hypernyms() == [w.synset("test-en-ext-0009-v")]
    assert w.synset("test-en-ext-0009-v").hyponyms() == [w.synset("test-en-0007-v")]
    assert w.sense("test-en-information-n-0001-01").get_related("pertainym") == [
        w.sense("test-en-ext-info-n-0001-01")
    ]
    assert w.sense("test-en-ext-info-n-0001-01").get_related("pertainym") == [
        w.sense("test-en-information-n-0001-01")
    ]

    # restricted to extension
    w = wn.Wordnet(lexicon="test-en-ext")
    assert w.synset("test-en-ext-0009-v").hyponyms() == []
    assert w.sense("test-en-ext-info-n-0001-01").get_related("pertainym") == []


@pytest.mark.usefixtures("mini_db_1_1")
def test_sense_synset_issue_168():
    ja = wn.Wordnet(lexicon="test-ja", expand="")
    assert ja.synset("test-ja-0001-n").get_related() == []
    assert ja.sense("test-ja-情報-n-0001-01").synset().get_related() == []



"""Web interface for Wn databases."""

from typing import Optional, Union
from functools import wraps
from urllib.parse import urlsplit, parse_qs, urlencode

from starlette.applications import Starlette  # type: ignore
from starlette.responses import JSONResponse  # type: ignore
from starlette.routing import Route  # type: ignore
from starlette.requests import Request  # type: ignore

import wn

DEFAULT_PAGINATION_LIMIT = 50


def paginate(proto):
    def paginate_wrapper(func):
        @wraps(func)
        async def _paginate_wrapper(request: Request) -> JSONResponse:
            url = str(request.url)
            query = dict(request.query_params)
            offset = abs(int(query.pop("page[offset]", 0)))
            limit = abs(int(query.pop("page[limit]", DEFAULT_PAGINATION_LIMIT)))

            obj = await func(request)
            total = len(obj["data"])
            prev = max(0, offset - limit)
            next = offset + limit
            last = (total // limit) * limit

            obj["data"] = [
                proto(x, request) for x in obj["data"][offset: offset + limit]
            ]
            obj.setdefault("meta", {}).update(total=total)

            links = {}
            if offset > 0:
                links["first"] = replace_query_params(url, **{"page[offset]": 0})
                links["prev"] = replace_query_params(url, **{"page[offset]": prev})
            if next < total:
                links["next"] = replace_query_params(url, **{"page[offset]": next})
                links["last"] = replace_query_params(url, **{"page[offset]": last})
            if links:
                obj.setdefault("links", {}).update(links)

            return JSONResponse(obj)

        return _paginate_wrapper

    return paginate_wrapper


def replace_query_params(url: str, **params) -> str:
    u = urlsplit(url)
    q = parse_qs(u.query)
    q.update(params)
    qs = urlencode(q, doseq=True)
    return u._replace(query=qs).geturl()


# Data-making functions


def _url_for_obj(
    request: Request,
    name: str,
    obj: Union[wn.Word, wn.Sense, wn.Synset],
    lexicon: Optional[str] = None,
) -> str:
    if lexicon is None:
        lexicon = obj.lexicon().specifier()
    kwargs = {"lexicon": lexicon, name: obj.id}
    return request.url_for(name, **kwargs)


def make_lexicon(lex: wn.Lexicon, request: Request) -> dict:
    spec = lex.specifier()
    return {
        "id": spec,
        "type": "lexicon",
        "attributes": {
            # cannot have 'id' as JSON:API disallows it
            "version": lex.version,
            "label": lex.label,
            "language": lex.language,
            "license": lex.license,
        },
        "links": {"self": request.url_for("lexicon", lexicon=spec)},
        "relationships": {
            "words": {
                "links": {"related": request.url_for("words", lexicon=spec)},
            },
            "synsets": {
                "links": {"related": request.url_for("synsets", lexicon=spec)},
            },
            "senses": {
                "links": {"related": request.url_for("senses", lexicon=spec)},
            },
        },
    }


def make_word(w: wn.Word, request: Request, basic: bool = False) -> dict:
    lex_spec = w.lexicon().specifier()
    d: dict = {
        "id": w.id,
        "type": "word",
        "attributes": {
            "pos": w.pos,
            "lemma": w.lemma(),
            "forms": w.forms(),
        },
        "links": {"self": _url_for_obj(request, "word", w, lexicon=lex_spec)},
    }
    if not basic:
        synsets = w.synsets()
        lex_link = request.url_for("lexicon", lexicon=lex_spec)
        senses_link = request.url_for("senses", word=w.id, lexicon=lex_spec)
        d.update(
            {
                "relationships": {
                    "senses": {"links": {"related": senses_link}},
                    "synsets": {
                        "data": [dict(type="synset", id=ss.id) for ss in synsets]
                    },
                    "lexicon": {"links": {"related": lex_link}},
                },
                "included": [make_synset(ss, request, basic=True) for ss in synsets],
            }
        )
    return d


def make_sense(s: wn.Sense, request: Request, basic: bool = False) -> dict:
    lex_spec = s.lexicon().specifier()
    d: dict = {
        "id": s.id,
        "type": "sense",
        "links": {"self": _url_for_obj(request, "sense", s, lexicon=lex_spec)},
    }
    if not basic:
        w = s.word()
        ss = s.synset()
        lex_link = request.url_for("lexicon", lexicon=lex_spec)
        word_link = request.url_for("word", word=w.id, lexicon=lex_spec)
        synset_link = request.url_for("synset", synset=ss.id, lexicon=lex_spec)
        relationships: dict = {
            "word": {"links": {"related": word_link}},
            "synset": {"links": {"related": synset_link}},
            "lexicon": {"links": {"related": lex_link}},
        }
        included = []
        for relname, slist in s.relations().items():
            relationships[relname] = {
                "data": [dict(type="sense", id=_s.id) for _s in slist]
            }
            included.extend([make_sense(_s, request, basic=True) for _s in slist])
        d.update({"relationships": relationships, "included": included})
    return d


def make_synset(ss: wn.Synset, request: Request, basic: bool = False) -> dict:
    lex_spec = ss.lexicon().specifier()
    d: dict = {
        "id": ss.id,
        "type": "synset",
        "attributes": {
            "pos": ss.pos,
            "ili": ss._ili,
        },
        "links": {"self": _url_for_obj(request, "synset", ss, lexicon=lex_spec)},
    }
    if not basic:
        words = ss.words()
        lex_link = request.url_for("lexicon", lexicon=lex_spec)
        members_link = request.url_for("senses", synset=ss.id, lexicon=lex_spec)
        relationships: dict = {
            "members": {"links": {"related": members_link}},
            "words": {"data": [dict(type="word", id=w.id) for w in words]},
            "lexicon": {"links": {"related": lex_link}},
        }
        included = [make_word(w, request, basic=True) for w in words]
        for relname, sslist in ss.relations().items():
            relationships[relname] = {
                "data": [dict(type="synset", id=_s.id) for _s in sslist]
            }
            included.extend([make_synset(_s, request, basic=True) for _s in sslist])
        d.update({"relationships": relationships, "included": included})
    return d


@paginate(make_lexicon)
async def lexicons(request):
    query = request.query_params
    _lexicons = wn.lexicons(
        lexicon=query.get("lexicon"),
        lang=query.get("lang"),
    )
    return {"data": _lexicons}


async def lexicon(request):
    path_params = request.path_params
    lex = wn.lexicons(lexicon=path_params["lexicon"])[0]
    return JSONResponse({"data": make_lexicon(lex, request)})


@paginate(make_word)
async def words(request):
    query = request.query_params
    lexicon = request.path_params.get("lexicon") or query.get("lexicon")
    _words = wn.words(
        form=query.get("form"),
        pos=query.get("pos"),
        lexicon=lexicon,
        lang=query.get("lang"),
    )
    return {"data": _words}


async def word(request):
    path_params = request.path_params
    word = wn.word(path_params["word"], lexicon=path_params["lexicon"])
    return JSONResponse({"data": make_word(word, request)})


@paginate(make_sense)
async def senses(request):
    query = request.query_params
    path = request.path_params
    lexicon = path.get("lexicon") or query.get("lexicon")
    if "word" in path:
        _senses = wn.word(path["word"], lexicon=lexicon).senses()
    elif "synset" in path:
        _senses = wn.synset(path["synset"], lexicon=lexicon).members()
    else:
        _senses = wn.senses(
            form=query.get("form"),
            pos=query.get("pos"),
            lexicon=lexicon,
            lang=query.get("lang"),
        )
    return {"data": _senses}


async def sense(request):
    path_params = request.path_params
    sense = wn.sense(path_params["sense"], lexicon=path_params["lexicon"])
    return JSONResponse({"data": make_sense(sense, request)})


@paginate(make_synset)
async def synsets(request):
    query = request.query_params
    lexicon = request.path_params.get("lexicon") or query.get("lexicon")
    _synsets = wn.synsets(
        form=query.get("form"),
        pos=query.get("pos"),
        ili=query.get("ili"),
        lexicon=lexicon,
        lang=query.get("lang"),
    )
    return {"data": _synsets}


async def synset(request):
    path_params = request.path_params
    synset = wn.synset(path_params["synset"], lexicon=path_params["lexicon"])
    return JSONResponse({"data": make_synset(synset, request)})


routes = [
    Route("/lexicons", endpoint=lexicons),
    Route("/lexicons/{lexicon}", endpoint=lexicon),
    Route("/lexicons/{lexicon}/words", endpoint=words),
    Route("/lexicons/{lexicon}/words/{word}", endpoint=word),
    Route("/lexicons/{lexicon}/words/{word}/senses", endpoint=senses),
    Route("/lexicons/{lexicon}/senses", endpoint=senses),
    Route("/lexicons/{lexicon}/senses/{sense}", endpoint=sense),
    Route("/lexicons/{lexicon}/synsets", endpoint=synsets),
    Route("/lexicons/{lexicon}/synsets/{synset}", endpoint=synset),
    Route("/lexicons/{lexicon}/synsets/{synset}/members", endpoint=senses),
    Route("/words", endpoint=words),
    Route("/senses", endpoint=senses),
    Route("/synsets", endpoint=synsets),
]

app = Starlette(debug=True, routes=routes)



"""Wn utility classes."""
from typing import TextIO, Callable
import sys


def synset_id_formatter(fmt: str = "{prefix}-{offset:08}-{pos}", **kwargs) -> Callable:
    """Return a function for formatting synset ids.

    The *fmt* argument can be customized. It will be formatted using
    any other keyword arguments given to this function and any given
    to the resulting function. By default, the format string expects a
    ``prefix`` string argument for the namespace (such as a lexicon
    id), an ``offset`` integer argument (such as a WNDB offset), and a
    ``pos`` string argument.

    Arguments:
        fmt: A Python format string
        **kwargs: Keyword arguments for the format string.

    Example:

        >>> pwn_synset_id = synset_id_formatter(prefix='pwn')
        >>> pwn_synset_id(offset=1174, pos='n')
        'pwn-00001174-n'

    """

    def format_synset_id(**_kwargs) -> str:
        return fmt.format(**kwargs, **_kwargs)

    return format_synset_id


class ProgressHandler:
    """An interface for updating progress in long-running processes.

    Long-running processes in Wn, such as :func:`wn.download` and
    :func:`wn.add`, call to a progress handler object as they go.  The
    default progress handler used by Wn is :class:`ProgressBar`, which
    updates progress by formatting and printing a textual bar to
    stderr. The :class:`ProgressHandler` class may be used directly,
    which does nothing, or users may create their own subclasses for,
    e.g., updating a GUI or some other handler.

    The initialization parameters, except for ``file``, are stored in
    a :attr:`kwargs` member and may be updated after the handler is
    created through the :meth:`set` method. The :meth:`update` method
    is the primary way a counter is updated. The :meth:`flash` method
    is sometimes called for simple messages. When the process is
    complete, the :meth:`close` method is called, optionally with a
    message.

    """

    def __init__(
        self,
        *,
        message: str = "",
        count: int = 0,
        total: int = 0,
        refresh_interval: int = 0,
        unit: str = "",
        status: str = "",
        file: TextIO = sys.stderr,
    ):
        self.file = file
        self.kwargs = {
            "count": count,
            "total": total,
            "refresh_interval": refresh_interval,
            "message": message,
            "unit": unit,
            "status": status,
        }
        self._refresh_quota: int = refresh_interval

    def update(self, n: int = 1, force: bool = False) -> None:
        """Update the counter with the increment value *n*.

        This method should update the ``count`` key of :attr:`kwargs`
        with the increment value *n*. After this, it is expected to
        update some user-facing progress indicator.

        If *force* is :python:`True`, any indicator will be refreshed
        regardless of the value of the refresh interval.

        """
        self.kwargs["count"] += n  # type: ignore

    def set(self, **kwargs) -> None:
        """Update progress handler parameters.

        Calling this method also runs :meth:`update` with an increment
        of 0, which causes a refresh of any indicator without changing
        the counter.

        """
        self.kwargs.update(**kwargs)
        self.update(0, force=True)

    def flash(self, message: str) -> None:
        """Issue a message unrelated to the current counter.

        This may be useful for multi-stage processes to indicate the
        move to a new stage, or to log unexpected situations.

        """
        pass

    def close(self) -> None:
        """Close the progress handler.

        This might be useful for closing file handles or cleaning up
        resources.

        """
        pass


class ProgressBar(ProgressHandler):
    """A :class:`ProgressHandler` subclass for printing a progress bar.

    Example:
        >>> p = ProgressBar(message='Progress: ', total=10, unit=' units')
        >>> p.update(3)
        Progress: [#########                     ] (3/10 units)

    See :meth:`format` for a description of how the progress bar is
    formatted.

    """

    #: The default formatting template.
    FMT = "\r{message}{bar}{counter}{status}"

    def update(self, n: int = 1, force: bool = False) -> None:
        """Increment the count by *n* and print the reformatted bar."""
        self.kwargs["count"] += n  # type: ignore
        self._refresh_quota -= n
        if force or self._refresh_quota <= 0:
            self._refresh_quota = self.kwargs["refresh_interval"]  # type: ignore
            s = self.format()
            if self.file:
                print("\r\033[K", end="", file=self.file)
                print(s, end="", file=self.file)

    def format(self) -> str:
        """Format and return the progress bar.

        The bar is is formatted according to :attr:`FMT`, using
        variables from :attr:`kwargs` and two computed variables:

        - ``bar``: visualization of the progress bar, empty when
          ``total`` is 0

        - ``counter``: display of ``count``, ``total``, and ``units``

        >>> p = ProgressBar(message='Progress', count=2, total=10, unit='K')
        >>> p.format()
        '\\rProgress [######                        ] (2/10K) '
        >>> p = ProgressBar(count=2, status='Counting...')
        >>> p.format()
        '\\r (2) Counting...'

        """
        _kw = self.kwargs
        width = 30
        total: int = _kw["total"]  # type: ignore
        count: int = _kw["count"]  # type: ignore

        if total > 0:
            num = min(count, total) * width
            fill = (num // total) * "#"
            part = ((num % total) * 3) // total
            if part:
                fill += "-="[part - 1]
            bar = f" [{fill:<{width}}]"
            counter = f' ({count}/{total}{_kw["unit"]}) '
        else:
            bar = ""
            counter = f' ({count}{_kw["unit"]}) '

        return self.FMT.format(bar=bar, counter=counter, **_kw)

    def flash(self, message: str) -> None:
        """Overwrite the progress bar with *message*."""
        print(f"\r\033[K{message}", end="", file=self.file)

    def close(self) -> None:
        """Print a newline so the last printed bar remains on screen."""
        print(file=self.file)



from typing import List, Dict, Set, Tuple, Sequence, Optional, cast

import wn
from wn._types import AnyPath, VersionInfo
from wn._util import version_info
from wn import lmf
from wn._queries import (
    find_entries,
    find_senses,
    find_synsets,
    find_syntactic_behaviours,
    find_proposed_ilis,
    get_entry_senses,
    get_sense_relations,
    get_sense_synset_relations,
    get_synset_relations,
    get_synset_members,
    get_examples,
    get_definitions,
    get_metadata,
    get_lexicalized,
    get_adjposition,
    get_form_pronunciations,
    get_form_tags,
    get_sense_counts,
    get_lexfile,
    get_lexicon,
    get_lexicon_dependencies,
    get_lexicon_extension_bases,
)
from wn._core import Lexicon


def export(
    lexicons: Sequence[Lexicon], destination: AnyPath, version: str = "1.0"
) -> None:
    """Export lexicons from the database to a WN-LMF file.

    More than one lexicon may be exported in the same file, subject to
    these conditions:

    - identifiers on wordnet entities must be unique in all lexicons
    - lexicons extensions may not be exported with their dependents

    >>> w = wn.Wordnet(lexicon='cmnwn zsmwn')
    >>> wn.export(w.lexicons(), 'cmn-zsm.xml')

    Args:
        lexicons: sequence of :class:`wn.Lexicon` objects
        destination: path to the destination file
        version: LMF version string

    """
    _precheck(lexicons)
    assert version in lmf.SUPPORTED_VERSIONS
    _version = version_info(version)
    resource: lmf.LexicalResource = {
        "lmf_version": version,
        "lexicons": [_export_lexicon(lex, _version) for lex in lexicons],
    }
    lmf.dump(resource, destination)


def _precheck(lexicons: Sequence[Lexicon]) -> None:
    all_ids: Set[str] = set()
    for lex in lexicons:
        lexids = (lex._id,)
        idset = {lex.id}
        idset.update(row[0] for row in find_entries(lexicon_rowids=lexids))
        idset.update(row[0] for row in find_senses(lexicon_rowids=lexids))
        idset.update(row[0] for row in find_synsets(lexicon_rowids=lexids))
        # TODO: syntactic behaviours
        if all_ids.intersection(idset):
            raise wn.Error("cannot export: non-unique identifiers in lexicons")
        all_ids |= idset


_SBMap = Dict[str, List[Tuple[str, str]]]


def _export_lexicon(lexicon: Lexicon, version: VersionInfo) -> lmf.Lexicon:
    lexids = (lexicon._id,)

    # WN-LMF 1.0 lexicons put syntactic behaviours on lexical entries
    # WN-LMF 1.1 lexicons use a 'subcat' IDREFS attribute
    sbmap: _SBMap = {}
    if version < (1, 1):
        for sbid, frame, sids in find_syntactic_behaviours(lexicon_rowids=lexids):
            for sid in sids:
                sbmap.setdefault(sid, []).append((sbid, frame))

    lex: lmf.Lexicon = {
        "id": lexicon.id,
        "label": lexicon.label,
        "language": lexicon.language,
        "email": lexicon.email,
        "license": lexicon.license,
        "version": lexicon.version,
        "url": lexicon.url or "",
        "citation": lexicon.citation or "",
        "entries": _export_lexical_entries(lexids, sbmap, version),
        "synsets": _export_synsets(lexids, version),
        "meta": _export_metadata(lexicon._id, "lexicons"),
    }
    if version >= (1, 1):
        lex["logo"] = lexicon.logo or ""
        lex["requires"] = _export_requires(lexicon._id)
        lex["frames"] = _export_syntactic_behaviours_1_1(lexids)

    return lex


def _export_requires(lexid: int) -> List[lmf.Dependency]:
    return [
        {"id": id, "version": version, "url": url}
        for id, version, url, _ in get_lexicon_dependencies(lexid)
    ]


def _export_extends(lexid: int) -> lmf.Dependency:
    ext_lexid = get_lexicon_extension_bases(lexid, depth=1)[0]
    _, id, _, _, _, _, version, url, *_ = get_lexicon(ext_lexid)
    return {"id": id, "version": version, "url": url}


def _export_lexical_entries(
    lexids: Sequence[int], sbmap: _SBMap, version: VersionInfo
) -> List[lmf.LexicalEntry]:
    entries: List[lmf.LexicalEntry] = []
    for id, pos, forms, _, rowid in find_entries(lexicon_rowids=lexids):
        entry: lmf.LexicalEntry = {
            "id": id,
            "lemma": {
                "writtenForm": forms[0][0],
                "partOfSpeech": pos,
                "script": forms[0][2] or "",
                "tags": _export_tags(forms[0][3]),
            },
            "forms": [],
            "senses": _export_senses(rowid, lexids, sbmap, version),
            "meta": _export_metadata(rowid, "entries"),
        }
        if version >= (1, 1):
            entry["lemma"]["pronunciations"] = _export_pronunciations(forms[0][3])
        for form, fid, script, frowid in forms[1:]:
            _form: lmf.Form = {
                "id": fid or "",
                "writtenForm": form,
                "script": script or "",
                "tags": _export_tags(frowid),
            }
            if version >= (1, 1):
                _form["pronunciations"] = _export_pronunciations(frowid)
            entry["forms"].append(_form)
        if version < (1, 1):
            entry["frames"] = _export_syntactic_behaviours_1_0(entry, sbmap)
        entries.append(entry)
    return entries


def _export_pronunciations(rowid: int) -> List[lmf.Pronunciation]:
    return [
        {
            "text": text,
            "variety": variety,
            "notation": notation,
            "phonemic": phonemic,
            "audio": audio,
        }
        for text, variety, notation, phonemic, audio in get_form_pronunciations(rowid)
    ]


def _export_tags(rowid: int) -> List[lmf.Tag]:
    return [
        {"text": text, "category": category} for text, category in get_form_tags(rowid)
    ]


def _export_senses(
    entry_rowid: int,
    lexids: Sequence[int],
    sbmap: _SBMap,
    version: VersionInfo,
) -> List[lmf.Sense]:
    senses: List[lmf.Sense] = []
    for id, _, synset, _, rowid in get_entry_senses(entry_rowid, lexids):
        sense: lmf.Sense = {
            "id": id,
            "synset": synset,
            "relations": _export_sense_relations(rowid, lexids),
            "examples": _export_examples(rowid, "senses", lexids),
            "counts": _export_counts(rowid, lexids),
            "lexicalized": get_lexicalized(rowid, "senses"),
            "adjposition": get_adjposition(rowid) or "",
            "meta": _export_metadata(rowid, "senses"),
        }
        if version >= (1, 1) and id in sbmap:
            sense["subcat"] = sorted(sbid for sbid, _ in sbmap[id])
        senses.append(sense)
    return senses


def _export_sense_relations(
    sense_rowid: int, lexids: Sequence[int]
) -> List[lmf.Relation]:
    relations: List[lmf.Relation] = [
        {
            "target": id,
            "relType": type,
            "meta": _export_metadata(rowid, "sense_relations"),
        }
        for type, rowid, id, *_ in get_sense_relations(sense_rowid, "*", lexids)
    ]
    relations.extend(
        {
            "target": id,
            "relType": type,
            "meta": _export_metadata(rowid, "sense_synset_relations"),
        }
        for type, rowid, id, *_ in get_sense_synset_relations(sense_rowid, "*", lexids)
    )
    return relations


def _export_examples(
    rowid: int, table: str, lexids: Sequence[int]
) -> List[lmf.Example]:
    return [
        {
            "text": text,
            "language": language,
            "meta": _export_metadata(rowid, f"{table[:-1]}_examples"),
        }
        for text, language, rowid in get_examples(rowid, table, lexids)
    ]


def _export_counts(rowid: int, lexids: Sequence[int]) -> List[lmf.Count]:
    return [
        {"value": val, "meta": _export_metadata(id, "counts")}
        for val, id in get_sense_counts(rowid, lexids)
    ]


def _export_synsets(lexids: Sequence[int], version: VersionInfo) -> List[lmf.Synset]:
    synsets: List[lmf.Synset] = []
    for id, pos, ili, _, rowid in find_synsets(lexicon_rowids=lexids):
        ilidef = _export_ili_definition(rowid)
        if ilidef and not ili:
            ili = "in"  # special case for proposed ILIs
        ss: lmf.Synset = {
            "id": id,
            "ili": ili or "",
            "partOfSpeech": pos,
            "definitions": _export_definitions(rowid, lexids),
            "relations": _export_synset_relations(rowid, lexids),
            "examples": _export_examples(rowid, "synsets", lexids),
            "lexicalized": get_lexicalized(rowid, "synsets"),
            "lexfile": get_lexfile(rowid) or "",
            "meta": _export_metadata(rowid, "synsets"),
        }
        if ilidef:
            ss["ili_definition"] = ilidef
        if version >= (1, 1):
            ss["members"] = [row[0] for row in get_synset_members(rowid, lexids)]
        synsets.append(ss)
    return synsets


def _export_definitions(rowid: int, lexids: Sequence[int]) -> List[lmf.Definition]:
    return [
        {
            "text": text,
            "language": language,
            "sourceSense": sense_id,
            "meta": _export_metadata(rowid, "definitions"),
        }
        for text, language, sense_id, rowid in get_definitions(rowid, lexids)
    ]


def _export_ili_definition(synset_rowid: int) -> Optional[lmf.ILIDefinition]:
    _, _, defn, rowid = next(
        find_proposed_ilis(synset_rowid=synset_rowid), (None, None, None, None)
    )
    ilidef: Optional[lmf.ILIDefinition] = None
    if defn:
        meta = None
        if rowid is not None:
            meta = _export_metadata(rowid, "proposed_ilis")
        ilidef = {"text": defn, "meta": meta}
    return ilidef


def _export_synset_relations(
    synset_rowid: int, lexids: Sequence[int]
) -> List[lmf.Relation]:
    return [
        {
            "target": id,
            "relType": type,
            "meta": _export_metadata(rowid, "synset_relations"),
        }
        for type, rowid, id, *_ in get_synset_relations((synset_rowid,), "*", lexids)
    ]


def _export_syntactic_behaviours_1_0(
    entry: lmf.LexicalEntry,
    sbmap: _SBMap,
) -> List[lmf.SyntacticBehaviour]:
    frames: List[lmf.SyntacticBehaviour] = []
    sense_ids = {s["id"] for s in entry.get("senses", [])}
    sbs: Dict[str, Set[str]] = {}
    for sid in sense_ids:
        for _, subcat_frame in sbmap.get(sid, []):
            sbs.setdefault(subcat_frame, set()).add(sid)
    for subcat_frame, sids in sbs.items():
        frame: lmf.SyntacticBehaviour = {
            "subcategorizationFrame": subcat_frame,
            "senses": sorted(sids),
        }
        frames.append(frame)
    return frames


def _export_syntactic_behaviours_1_1(
    lexids: Sequence[int],
) -> List[lmf.SyntacticBehaviour]:
    return [
        {"id": id or "", "subcategorizationFrame": frame}
        for id, frame, _ in find_syntactic_behaviours(lexicon_rowids=lexids)
    ]


def _export_metadata(rowid: int, table: str) -> lmf.Metadata:
    return cast(lmf.Metadata, get_metadata(rowid, table))



import sqlite3
import threading
import tempfile

import pytest

import wn


@pytest.mark.usefixtures("mini_db")
def test_schema_compatibility():
    conn = sqlite3.connect(str(wn.config.database_path))
    schema_hash = wn._db.schema_hash(conn)
    assert schema_hash in wn._db.COMPATIBLE_SCHEMA_HASHES


@pytest.mark.usefixtures("mini_db")
def test_db_multithreading():
    """
    See https://github.com/goodmami/wn/issues/86
    Thanks: @fushinari
    """

    class WNThread:
        w = None

        def __init__(self):
            w_thread = threading.Thread(target=self.set_w)
            w_thread.start()
            w_thread.join()
            self.w.synsets()

        def set_w(self):
            if self.w is None:
                self.w = wn.Wordnet()

    # close the connections by resetting the pool
    wn._db.pool = {}
    with pytest.raises(sqlite3.ProgrammingError):
        WNThread()
    wn._db.pool = {}
    wn.config.allow_multithreading = True
    WNThread()  # no error
    wn.config.allow_multithreading = False
    wn._db.pool = {}


def test_remove_extension(mini_lmf_1_0, mini_lmf_1_1):
    with tempfile.TemporaryDirectory("wn_data_1_1_trigger") as dir:
        old_data_dir = wn.config.data_directory
        wn.config.data_directory = dir
        wn.add(mini_lmf_1_0)
        wn.add(mini_lmf_1_1)
        assert len(wn.lexicons()) == 4
        wn.remove("test-en-ext")
        assert len(wn.lexicons()) == 3
        wn.remove("test-ja")
        assert len(wn.lexicons()) == 2
        wn.add(mini_lmf_1_1)
        assert len(wn.lexicons()) == 4
        wn.remove("test-en")
        assert {lex.id for lex in wn.lexicons()} == {"test-es", "test-ja"}
        wn.config.data_directory = old_data_dir
        # close any open DB connections before teardown
        for conn in wn._db.pool.values():
            conn.close()



"""Non-public Wn utilities."""

from typing import TypeVar, Iterable, List
import sys
from pathlib import Path
import hashlib
from unicodedata import normalize, combining

# version check is for mypy; see https://github.com/python/mypy/issues/1153
if sys.version_info >= (3, 7):
    import importlib.resources as resources
else:
    import importlib_resources as resources  # noqa: F401


from wn._types import VersionInfo


def version_info(version_string: str) -> VersionInfo:
    return tuple(map(int, version_string.split(".")))


def is_url(string: str) -> bool:
    """Return True if *string* appears to be a URL."""
    # TODO: ETags?
    return any(string.startswith(scheme) for scheme in ("http://", "https://"))


def is_gzip(path: Path) -> bool:
    """Return True if the file at *path* appears to be gzipped."""
    return _inspect_file_signature(path, b"\x1F\x8B")


def is_lzma(path: Path) -> bool:
    """Return True if the file at *path* appears to be lzma-compressed."""
    return _inspect_file_signature(path, b"\xFD7zXZ\x00")


def is_xml(path: Path) -> bool:
    """Return True if the file at *path* appears to be an XML file."""
    return _inspect_file_signature(path, b"<?xml ")


def _inspect_file_signature(path: Path, signature: bytes) -> bool:
    if path.is_file():
        with path.open("rb") as f:
            return f.read(len(signature)) == signature
    return False


def short_hash(string: str) -> str:
    """Return a short hash of *string*."""
    b2 = hashlib.blake2b(digest_size=20)
    b2.update(string.encode("utf-8"))
    return b2.hexdigest()


T = TypeVar("T")


def flatten(iterable: Iterable[Iterable[T]]) -> List[T]:
    return [x for xs in iterable for x in xs]


def normalize_form(s: str) -> str:
    return "".join(c for c in normalize("NFKD", s.lower()) if not combining(c))
